@@ -3,10 +3,9 @@ Utilities for app components provided as part of the trulens package.
 Currently organizes all such components as "Other".
 """
 
-from typing import Type
-
 from trulens.core import app
-from trulens.core.utils.pyschema import Class
+from trulens.core._utils.pycompat import Type
+from trulens.core.utils import pyschema as pyschema_utils
 
 
 class Other(app.Other, app.TrulensComponent):
@@ -17,7 +16,9 @@ class Other(app.Other, app.TrulensComponent):
 COMPONENT_VIEWS = [Other]
 
 
-def constructor_of_class(cls_obj: Class) -> Type[app.TrulensComponent]:
+def constructor_of_class(
+    cls_obj: pyschema_utils.Class,
+) -> Type[app.TrulensComponent]:
     for view in COMPONENT_VIEWS:
         if view.class_is(cls_obj):
             return view
@@ -26,7 +27,7 @@ def constructor_of_class(cls_obj: Class) -> Type[app.TrulensComponent]:
 
 
 def component_of_json(json: dict) -> app.TrulensComponent:
-    cls = Class.of_class_info(json)
+    cls = pyschema_utils.Class.of_class_info(json)
 
     view = constructor_of_class(cls)
 
